include_parents = False
recursive_queries = True
clever_parsing = True
query_exceptions = True
native_types = False  # if True, the result of get method, or first, last, etc will be a python object instead of a jsonnode
